#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>

#include "bitstream.h"

#define BUFFER_SIZE 32768
#define BITMASK(x) ((1 << (x)) - 1)

int bitmask[32];

u_char read_byte(bitstream *b) {
  if(b->buf_pos < BUFFER_SIZE) return b->buffer[b->buf_pos++];
  else if(read(b->fd, b->buffer, BUFFER_SIZE) <= 0) {
    fprintf(stderr, "End of file, quitting\n");
    exit(0);
  }
  //  b->filepos += 256;
  b->buf_pos = 1;
  return(*(b->buffer));
}

unsigned long read_bits(bitstream *b, int num) {
  while (b->num_bits < num) {
    b->num_bits += 8;
    b->cur = (read_byte(b) | (b->cur << 8));
  }
  b->num_bits -= num;
  //return((b->cur >> b->num_bits) & BITMASK(num));
   return((b->cur >> b->num_bits) & bitmask[num]);
}

unsigned long peek_bits(bitstream *b, int num) {
  while (b->num_bits < num) {
    b->num_bits += 8;
    b->cur = (read_byte(b) | (b->cur << 8));
  }
  return((b->cur >> (b->num_bits - num)) & BITMASK(num));
}

void align_bits(bitstream *b) {
  b->num_bits -= b->num_bits % 8;
}


bitstream *init_bitstream(const char *fname) {
  bitstream *bs;
  int i;

  bs = (bitstream *)malloc(sizeof(bitstream));
  bs->buffer = (u_char *)malloc(BUFFER_SIZE);
  bs->buf_pos = 0;
  bs->num_bits = 0;
  bs->cur = 0;
  bs->fd = open(fname, O_RDONLY);
  if(bs->fd < 0) {
    perror("init_bitstream");
    exit(-1);
  }

  for(i=0; i<32; i++)
    bitmask[i] = BITMASK(i);

  read(bs->fd, bs->buffer, BUFFER_SIZE);
  //  bs->filepos = 256;
  return bs;
}
