
#include "fast_1d_dct.h"

int dct(int pels[8], float coefs[8]) {
  static int i;
  static float c0, c1, c2, c3, c4, c5, c6, c7, c8 ,c9, c10, c11, c12;
  static float alpha;
  static float ts, m1, m2, m3;
  static float s1[8], s2[8], s3[8], s4[8], s5[8];
  static int first=1;
  double cos(), sin(), sqrt();

  if(first){
    first = 0;
    alpha = 2*PI/(4*8);

    c0 = sin(4.0*alpha) - cos(4.0*alpha);
    c1 = cos(4.0*alpha);
    c2 = -(sin(4.0*alpha) + cos(4.0*alpha));

    c3 = sin(1.0*alpha) - cos(1.0*alpha);
    c4 = cos(1.0*alpha);
    c5 = -(sin(1.0*alpha) + cos(1.0*alpha));

    c6 = sin(2.0*alpha) - cos(2.0*alpha);
    c7 = cos(2.0*alpha);
    c8 = -(sin(2.0*alpha) + cos(2.0*alpha));

    c9 = sin(3.0*alpha) - cos(3.0*alpha);
    c10 = cos(3.0*alpha);
    c11 = -(sin(3.0*alpha) + cos(3.0*alpha));

    c12 = (float) 1/ sqrt((float)2);
  }

  s1[0] = pels[0] + pels[7]; s1[1] = pels[0] - pels[7];
  s1[2] = pels[2] + pels[1]; s1[3] = pels[2] - pels[1];
  s1[4] = pels[4] + pels[3]; s1[5] = pels[4] - pels[3];
  s1[6] = pels[5] + pels[6]; s1[7] = pels[5] - pels[6];

  s2[0] = s1[1]; s2[1] = s1[5];
  s2[2] = s1[0] + s1[4]; s2[3] = s1[0] - s1[4];
  s2[4] = s1[2] + s1[6]; s2[5] = s1[2] - s1[6];
  s2[6] = s1[3] + s1[7]; s2[7] = s1[3] - s1[7];

  s3[0] = s2[0]; s3[1] = s2[1];
  s3[2] = s2[2]; s3[3] = s2[5]*c12;
  s3[4] = s2[3]; s3[5] = s2[4];
  s3[6] = s2[6]; s3[7] = s2[7]*c12;

  /* The following can be incorporated into the quantization step */

  for(i = 0; i < 8; ++i) s3[i] = s3[i]/2.0;

  s4[0] = s3[2]; s4[1] = s3[4];
  s4[2] = s3[5]; s4[3] = s3[6];
  s4[4] = s3[0] + s3[3]; s4[5] = s3[0] - s3[3];
  s4[6] = s3[7] + s3[1]; s4[7] = s3[7] - s3[1];

  ts = s4[0] + s4[2];
  m1 = s4[0]*c0;
  m2 = ts*c1;
  m3 = s4[2]*c2;
  s5[0] = m2 + m3;
  s5[1] = m1 + m2;
  
  ts = s4[4] + s4[6];
  m1 = s4[4]*c3;
  m2 = ts*c4;
  m3 = s4[6]*c5;
  s5[2] = m2 + m3;
  s5[3] = m1 + m2;
  
  ts = s4[1] + s4[3];
  m1 = s4[1]*c6;
  m2 = ts*c7;
  m3 = s4[3]*c8;
  s5[4] = m2 + m3;
  s5[5] = m1 + m2;
  
  ts = s4[5] + s4[7];
  m1 = s4[5]*c9;
  m2 = ts*c10;
  m3 = s4[7]*c11;
  s5[6] = m2 + m3;
  s5[7] = m1 + m2;
  
  coefs[0] = s5[1];
  coefs[1] = s5[2];
  coefs[2] = s5[4];
  coefs[3] = s5[6];
  coefs[4] = s5[0];
  coefs[5] = s5[7];
  coefs[6] = s5[5];
  coefs[7] = s5[3];

  return(0);
}

int idct(float coefs[8], int pels[8]){
  static float dout[8];
  static float c1, c2, c3, c4, c5, c6, c7, c8, c9;
  static float s1[8], s2[8], s3[8], s4[8], s5[8];
  static float nmax2, alpha;
  static int i;
  static int first = 1;
  
  if(first) {
    first = 0;
    nmax2 = 4;
    alpha = 2*PI/(4*8);
    
    c1 = cos(nmax2*alpha);   c2 = sin(nmax2*alpha);
    c3 = cos(alpha);         c4 = sin(alpha);
    c5 = cos(2*alpha);       c6 = sin(2*alpha);
    c7 = cos(3*alpha);       c8 = sin(3*alpha);
    c9 = (float) sqrt((float)2);
  }
   
  s1[0] = coefs[4];
  s1[1] = coefs[0];
  s1[2] = coefs[1];
  s1[3] = coefs[7];
  s1[4] = coefs[2];
  s1[5] = coefs[6];
  s1[6] = coefs[3];
  s1[7] = coefs[5];

  s2[0] = c1*2*s1[0] + c2*2*s1[1];  s2[1] = c5*2*s1[4] + c6*2*s1[5];
  s2[2] = c1*2*s1[1] - c2*2*s1[0];  s2[3] = c5*2*s1[5] - c6*2*s1[4];
  s2[4] = c3*2*s1[2] + c4*2*s1[3];  s2[5] = c7*2*s1[6] + c8*2*s1[7];   /* Can remove *2 and remove *0.5 below..??*/
  s2[6] = c3*2*s1[3] - c4*2*s1[2];  s2[7] = c7*2*s1[7] - c8*2*s1[6];

  s3[0] = 0.5 * s2[4] + 0.5 * s2[5];
  s3[1] = 0.5 * s2[6] - 0.5 * s2[7];
  s3[2] = s2[0];
  s3[3] = 0.5 * s2[4] - 0.5 * s2[5];
  s3[4] = s2[1];
  s3[5] = s2[2];
  s3[6] = s2[3];
  s3[7] = 0.5 * s2[6] + 0.5 * s2[7];

  s4[0] = s3[0];
  s4[1] = s3[1];
  s4[2] = s3[2];
  s4[3] = s3[4];
  s4[4] = s3[5];
  s4[5] = s3[3] * c9;
  s4[6] = s3[6];
  s4[7] = s3[7] * c9;

  s5[0] = 0.5 * s4[2] + 0.5 * s4[3];
  s5[1] = s4[0];
  s5[2] = 0.5 * s4[4] + 0.5 * s4[5];
  s5[3] = 0.5 * s4[6] + 0.5 * s4[7];
  s5[4] = 0.5 * s4[2] - 0.5 * s4[3];
  s5[5] = s4[1];
  s5[6] = 0.5 * s4[4] - 0.5 * s4[5];
  s5[7] = 0.5 * s4[6] - 0.5 * s4[7];

  dout[0] = 0.5 * s5[0] + 0.5 * s5[1];
  dout[1] = 0.5 * s5[2] - 0.5 * s5[3];
  dout[2] = 0.5 * s5[2] + 0.5 * s5[3];
  dout[3] = 0.5 * s5[4] - 0.5 * s5[5];
  dout[4] = 0.5 * s5[4] + 0.5 * s5[5];
  dout[5] = 0.5 * s5[6] + 0.5 * s5[7];
  dout[6] = 0.5 * s5[6] - 0.5 * s5[7];
  dout[7] = 0.5 * s5[0] - 0.5 * s5[1];

  for(i=0; i<=7; i++) {
    /*    if(dout[i] < 0.0) pels[i] = dout[i] - 0.5;
    else pels[i] = dout[i] + 0.5; */
    pels[i] = floor(dout[i] + 0.5);
  }
  
  /*
  if(dout[0] < 0.0) pels[0] = dout[0] - 0.5;
  else pels[0] = dout[0] + 0.5;
  if(dout[1] < 0.0) pels[1] = dout[1] - 0.5;
  else pels[1] = dout[1] + 0.5;
  if(dout[2] < 0.0) pels[2] = dout[2] - 0.5;
  else pels[2] = dout[2] + 0.5;
  if(dout[3] < 0.0) pels[3] = dout[3] - 0.5;
  else pels[3] = dout[3] + 0.5;
  if(dout[4] < 0.0) pels[4] = dout[4] - 0.5;
  else pels[4] = dout[4] + 0.5;
  if(dout[5] < 0.0) pels[5] = dout[5] - 0.5;
  else pels[5] = dout[5] + 0.5;
  if(dout[6] < 0.0) pels[6] = dout[6] - 0.5;
  else pels[6] = dout[6] + 0.5;
  if(dout[7] < 0.0) pels[7] = dout[7] - 0.5;
  else pels[7] = dout[7] + 0.5;
  */
  return(0);
}


int dct_f(float pels[8], float coefs[8]) {
  static int i;
  static float c0, c1, c2, c3, c4, c5, c6, c7, c8 ,c9, c10, c11, c12;
  static float alpha;
  static float ts, m1, m2, m3;
  static float s1[8], s2[8], s3[8], s4[8], s5[8];
  static int first=1;
  double cos(), sin(), sqrt();

  if(first){
    first = 0;
    alpha = 2*PI/(4*8);

    c0 = sin(4.0*alpha) - cos(4.0*alpha);
    c1 = cos(4.0*alpha);
    c2 = -(sin(4.0*alpha) + cos(4.0*alpha));

    c3 = sin(1.0*alpha) - cos(1.0*alpha);
    c4 = cos(1.0*alpha);
    c5 = -(sin(1.0*alpha) + cos(1.0*alpha));

    c6 = sin(2.0*alpha) - cos(2.0*alpha);
    c7 = cos(2.0*alpha);
    c8 = -(sin(2.0*alpha) + cos(2.0*alpha));

    c9 = sin(3.0*alpha) - cos(3.0*alpha);
    c10 = cos(3.0*alpha);
    c11 = -(sin(3.0*alpha) + cos(3.0*alpha));

    c12 = (float) 1/ sqrt((float)2);
  }
  
  s1[0] = pels[0] + pels[7]; s1[1] = pels[0] - pels[7];
  s1[2] = pels[2] + pels[1]; s1[3] = pels[2] - pels[1];
  s1[4] = pels[4] + pels[3]; s1[5] = pels[4] - pels[3];
  s1[6] = pels[5] + pels[6]; s1[7] = pels[5] - pels[6];

  s2[0] = s1[1]; s2[1] = s1[5];
  s2[2] = s1[0] + s1[4]; s2[3] = s1[0] - s1[4];
  s2[4] = s1[2] + s1[6]; s2[5] = s1[2] - s1[6];
  s2[6] = s1[3] + s1[7]; s2[7] = s1[3] - s1[7];

  s3[0] = s2[0]; s3[1] = s2[1];
  s3[2] = s2[2]; s3[3] = s2[5]*c12;
  s3[4] = s2[3]; s3[5] = s2[4];
  s3[6] = s2[6]; s3[7] = s2[7]*c12;

  /* The following can be incorporated into the quantization step */

  for(i = 0; i < 8; ++i) s3[i] = s3[i]/2.0;

  s4[0] = s3[2]; s4[1] = s3[4];
  s4[2] = s3[5]; s4[3] = s3[6];
  s4[4] = s3[0] + s3[3]; s4[5] = s3[0] - s3[3];
  s4[6] = s3[7] + s3[1]; s4[7] = s3[7] - s3[1];

  ts = s4[0] + s4[2];
  m1 = s4[0]*c0;
  m2 = ts*c1;
  m3 = s4[2]*c2;
  s5[0] = m2 + m3;
  s5[1] = m1 + m2;
  
  ts = s4[4] + s4[6];
  m1 = s4[4]*c3;
  m2 = ts*c4;
  m3 = s4[6]*c5;
  s5[2] = m2 + m3;
  s5[3] = m1 + m2;
  
  ts = s4[1] + s4[3];
  m1 = s4[1]*c6;
  m2 = ts*c7;
  m3 = s4[3]*c8;
  s5[4] = m2 + m3;
  s5[5] = m1 + m2;
  
  ts = s4[5] + s4[7];
  m1 = s4[5]*c9;
  m2 = ts*c10;
  m3 = s4[7]*c11;
  s5[6] = m2 + m3;
  s5[7] = m1 + m2;
  
  coefs[0] = s5[1];
  coefs[1] = s5[2];
  coefs[2] = s5[4];
  coefs[3] = s5[6];
  coefs[4] = s5[0];
  coefs[5] = s5[7];
  coefs[6] = s5[5];
  coefs[7] = s5[3];

  return(0);
}

int idct_f(float coefs[8], float pels[8]){
  static float dout[8];
  static float c1, c2, c3, c4, c5, c6, c7, c8, c9;
  static float s1[8], s2[8], s3[8], s4[8], s5[8];
  static float nmax2, alpha;
  static int i;
  static int first = 1;
  
  if(first) {
    first = 0;
    nmax2 = 4;
    alpha = 2*PI/(4*8);
    
    c1 = cos(nmax2*alpha);   c2 = sin(nmax2*alpha);
    c3 = cos(alpha);         c4 = sin(alpha);
    c5 = cos(2*alpha);       c6 = sin(2*alpha);
    c7 = cos(3*alpha);       c8 = sin(3*alpha);
    c9 = sqrt((float)2);
  }
   
  s1[0] = coefs[4];
  s1[1] = coefs[0];
  s1[2] = coefs[1];
  s1[3] = coefs[7];
  s1[4] = coefs[2];
  s1[5] = coefs[6];
  s1[6] = coefs[3];
  s1[7] = coefs[5];
  
  s2[0] = c1*2*s1[0] + c2*2*s1[1];  s2[1] = c5*2*s1[4] + c6*2*s1[5];
  s2[2] = c1*2*s1[1] - c2*2*s1[0];  s2[3] = c5*2*s1[5] - c6*2*s1[4];
  s2[4] = c3*2*s1[2] + c4*2*s1[3];  s2[5] = c7*2*s1[6] + c8*2*s1[7];   /* Can remove *2 and remove *0.5 below..??*/
  s2[6] = c3*2*s1[3] - c4*2*s1[2];  s2[7] = c7*2*s1[7] - c8*2*s1[6];
  
  s3[0] = 0.5 * s2[4] + 0.5 * s2[5];
  s3[1] = 0.5 * s2[6] - 0.5 * s2[7];
  s3[2] = s2[0];
  s3[3] = 0.5 * s2[4] - 0.5 * s2[5];
  s3[4] = s2[1];
  s3[5] = s2[2];
  s3[6] = s2[3];
  s3[7] = 0.5 * s2[6] + 0.5 * s2[7];

  s4[0] = s3[0];
  s4[1] = s3[1];
  s4[2] = s3[2];
  s4[3] = s3[4];
  s4[4] = s3[5];
  s4[5] = s3[3] * c9;
  s4[6] = s3[6];
  s4[7] = s3[7] * c9;

  s5[0] = 0.5 * s4[2] + 0.5 * s4[3];
  s5[1] = s4[0];
  s5[2] = 0.5 * s4[4] + 0.5 * s4[5];
  s5[3] = 0.5 * s4[6] + 0.5 * s4[7];
  s5[4] = 0.5 * s4[2] - 0.5 * s4[3];
  s5[5] = s4[1];
  s5[6] = 0.5 * s4[4] - 0.5 * s4[5];
  s5[7] = 0.5 * s4[6] - 0.5 * s4[7];

  dout[0] = 0.5 * s5[0] + 0.5 * s5[1];
  dout[1] = 0.5 * s5[2] - 0.5 * s5[3];
  dout[2] = 0.5 * s5[2] + 0.5 * s5[3];
  dout[3] = 0.5 * s5[4] - 0.5 * s5[5];
  dout[4] = 0.5 * s5[4] + 0.5 * s5[5];
  dout[5] = 0.5 * s5[6] + 0.5 * s5[7];
  dout[6] = 0.5 * s5[6] - 0.5 * s5[7];
  dout[7] = 0.5 * s5[0] - 0.5 * s5[1];

  /* ??? */
  
  for(i=0; i<=7; i++) {
    /* if(dout[i] < 0.0) pels[i] = dout[i] - 0.5;
    else pels[i] = dout[i] + 0.5;
    */

    pels[i] = dout[i];
  }
  

  return(0);
}



