#include "fast_2d_dct.h"

int dct2d(int pels[64], int coeffs[64]){
  static float tempcoeffs[64], moretempcoeffs[64];
  static int offset, tmp;

  dct(&pels[0], &tempcoeffs[0]);
  dct(&pels[8], &tempcoeffs[8]);
  dct(&pels[16], &tempcoeffs[16]);
  dct(&pels[24], &tempcoeffs[24]);
  dct(&pels[32], &tempcoeffs[32]);
  dct(&pels[40], &tempcoeffs[40]);  
  dct(&pels[48], &tempcoeffs[48]);
  dct(&pels[56], &tempcoeffs[56]);

  /* Transpose the memory... */
  for(offset = 0; offset <= 7; offset++) {
    moretempcoeffs[offset*8] = tempcoeffs[offset];
    moretempcoeffs[offset*8+1] = tempcoeffs[1*8 + offset];
    moretempcoeffs[offset*8+2] = tempcoeffs[2*8 + offset];
    moretempcoeffs[offset*8+3] = tempcoeffs[3*8 + offset];
    moretempcoeffs[offset*8+4] = tempcoeffs[4*8 + offset];
    moretempcoeffs[offset*8+5] = tempcoeffs[5*8 + offset];
    moretempcoeffs[offset*8+6] = tempcoeffs[6*8 + offset];
    moretempcoeffs[offset*8+7] = tempcoeffs[7*8 + offset];
  }

  dct_f(&moretempcoeffs[0], &tempcoeffs[0]);
  dct_f(&moretempcoeffs[8], &tempcoeffs[8]);
  dct_f(&moretempcoeffs[16], &tempcoeffs[16]);
  dct_f(&moretempcoeffs[24], &tempcoeffs[24]);
  dct_f(&moretempcoeffs[32], &tempcoeffs[32]);
  dct_f(&moretempcoeffs[40], &tempcoeffs[40]);
  dct_f(&moretempcoeffs[48], &tempcoeffs[48]);
  dct_f(&moretempcoeffs[56], &tempcoeffs[56]);

  /* Transpose back... and cast*/
  for(offset = 0; offset <= 7; offset++) {
    tmp = offset*8;
    coeffs[tmp] = (int) tempcoeffs[offset];
    coeffs[tmp+1] = (int) tempcoeffs[1*8 + offset];
    coeffs[tmp+2] = (int) tempcoeffs[2*8 + offset];
    coeffs[tmp+3] = (int) tempcoeffs[3*8 + offset];
    coeffs[tmp+4] = (int) tempcoeffs[4*8 + offset];
    coeffs[tmp+5] = (int) tempcoeffs[5*8 + offset];
    coeffs[tmp+6] = (int) tempcoeffs[6*8 + offset];
    coeffs[tmp+7] = (int) tempcoeffs[7*8 + offset];
  }

  return(0);
}

int idct2d(int coeffs[64], int pels[64]){
  static float tempcoeffs[64], moretempcoeffs[64];
  static int offset;
  int i;

  /* Transpose back... */
  for(offset = 0; offset <= 7; offset++) {
    tempcoeffs[offset*8] = (float) coeffs[offset];
    tempcoeffs[offset*8+1] = (float) coeffs[1*8 + offset];
    tempcoeffs[offset*8+2] = (float) coeffs[2*8 + offset];
    tempcoeffs[offset*8+3] = (float) coeffs[3*8 + offset];
    tempcoeffs[offset*8+4] = (float) coeffs[4*8 + offset];
    tempcoeffs[offset*8+5] = (float) coeffs[5*8 + offset];
    tempcoeffs[offset*8+6] = (float) coeffs[6*8 + offset];
    tempcoeffs[offset*8+7] = (float) coeffs[7*8 + offset];
  }

  idct_f(&tempcoeffs[0], &moretempcoeffs[0]);
  idct_f(&tempcoeffs[8], &moretempcoeffs[8]);
  idct_f(&tempcoeffs[16], &moretempcoeffs[16]);
  idct_f(&tempcoeffs[24], &moretempcoeffs[24]);
  idct_f(&tempcoeffs[32], &moretempcoeffs[32]);
  idct_f(&tempcoeffs[40], &moretempcoeffs[40]);  
  idct_f(&tempcoeffs[48], &moretempcoeffs[48]);
  idct_f(&tempcoeffs[56], &moretempcoeffs[56]);
  
  /* Transpose the memory... */
  for(offset = 0; offset <= 7; offset++) {
    tempcoeffs[offset*8] = moretempcoeffs[offset];
    tempcoeffs[offset*8+1] = moretempcoeffs[1*8 + offset];
    tempcoeffs[offset*8+2] = moretempcoeffs[2*8 + offset];
    tempcoeffs[offset*8+3] = moretempcoeffs[3*8 + offset];
    tempcoeffs[offset*8+4] = moretempcoeffs[4*8 + offset];
    tempcoeffs[offset*8+5] = moretempcoeffs[5*8 + offset];
    tempcoeffs[offset*8+6] = moretempcoeffs[6*8 + offset];
    tempcoeffs[offset*8+7] = moretempcoeffs[7*8 + offset];
  }

  idct(&tempcoeffs[0], &pels[0]);
  idct(&tempcoeffs[8], &pels[8]);
  idct(&tempcoeffs[16], &pels[16]);
  idct(&tempcoeffs[24], &pels[24]);
  idct(&tempcoeffs[32], &pels[32]);
  idct(&tempcoeffs[40], &pels[40]);
  idct(&tempcoeffs[48], &pels[48]);
  idct(&tempcoeffs[56], &pels[56]);

  for(i=0; i<64; i++) {
    if(pels[i] < -256) pels[i] = -256;
    if(pels[i] > 255) pels[i] = 255;
  }
  return(0);
}







