/****************************************************************
 * Peter Boettcher <pwb@andrew.cmu.edu>
 * 18-796 Homework #6
 * 3/30/99
 *
 * mmx.h: Hand-written MMX assembly for motion estimator
 ****************************************************************/

#ifndef MMX_H
#define MMX_H

/* Clear the accumulation register */
#define MMX_CLR __asm__("pxor %mm3,%mm3\npxor %mm7,%mm7")

/* Subtract two 8-byte buffers of data, byte for byte,
   take absolute value, and save in accumulation register */
/* MMX_SUB16 interleaves two loops of 8-byte blocks, to optimize
   pipeline usage.  Each loop uses independent registers, and
   instructions are placed in pairs which can be executed
   simultaneously.  This gives a very modest speed gain (from
   6.25 s to 6.06 */

#define MMX_SUB16(data1, data2, data3, data4) \
__asm__ volatile("\
movq %1,%%mm1 \n\
movq %0,%%mm0 \n\
movq %%mm1,%%mm2 \n\
psubusb %%mm0,%%mm1 \n\
movq %2,%%mm4 \n\
psubusb %%mm2,%%mm0 \n\
movq %3,%%mm5 \n\
por %%mm1,%%mm0 \n\
movq %%mm5,%%mm6 \n\
movq %%mm0,%%mm1 \n\
psubusb %%mm4,%%mm5 \n\
punpcklbw %%mm3, %%mm0 \n\
psubusb %%mm6,%%mm4 \n\
punpckhbw %%mm3, %%mm1 \n\
por %%mm5,%%mm4 \n\
paddusw %%mm0, %%mm7 \n\
movq %%mm4,%%mm5 \n\
paddusw %%mm1, %%mm7 \n\
punpcklbw %%mm3, %%mm4 \n\
punpckhbw %%mm3, %%mm5 \n\
paddusw %%mm4, %%mm7 \n\
paddusw %%mm5, %%mm7" \
	:  \
	: "m" (data1), "m" (data2), "m" (data3), "m" (data4))


/* Add up accumulation register and store it
   back to variable */
#define MMX_ADDALL(result) \
__asm__ volatile("\
movq %%mm7, %%mm0 \n\
psrlq $32, %%mm0\n\
paddusw %%mm7, %%mm0 \n\
movq %%mm0, %%mm7 \n\
psrlq $16, %%mm0 \n\
paddusw %%mm7, %%mm0 \n\
movq %%mm0, %0" \
	: "=m" (result))

/* You have to clean up when you're done with MMX */
#define MMX_EXIT __asm__("emms");

#endif /* MMX_H */
