
#include "parse.h"


extern int printflag;
extern motion_vec_state *mv_st;

//#define DEBUG
//#define EX_BITSTREAM

void parse_slice(bitstream *b, u_char slicenum);
void parse_macroblock(bitstream *b, pic_header *pic, int quant);
void init_mb();

pic_header *pic;

/* Look for three aligned zeros followed by a one */
char find_start_code(bitstream *b) {
  u_char c;

  align_bits(b);

 NO_ZEROS:
  if(read_bits(b, 8) == 0) goto ONE_ZERO;
  else goto NO_ZEROS;

 ONE_ZERO:
  if(read_bits(b, 8) == 0) goto TWO_ZEROS;
  else goto NO_ZEROS;

 TWO_ZEROS:
  if((c=read_bits(b, 8)) == 1) goto CODE_FOUND;
  else if(c == 0) goto TWO_ZEROS;
  else goto NO_ZEROS;

 CODE_FOUND:
  return(read_bits(b, 8));
}

seq_header *parse_sequence(bitstream *b) {
  seq_header *seq;
  int i;

  seq = (seq_header *)malloc(sizeof(seq_header));
  seq->horizontal=read_bits(b,12);
  seq->vertical=read_bits(b,12);
  seq->aspect=read_bits(b,4);
  seq->pic_rate=read_bits(b,4);
  seq->bit_rate=read_bits(b,18);
  assert(read_bits(b,1));
  seq->vbv_bufsize=read_bits(b,10);
  seq->constrained=read_bits(b,1);
  if(read_bits(b,1)) {
    fprintf(stderr, "Load intra quant matrix not implemented\n");
    for(i=0; i<16; i++)
      read_bits(b,32);
  }
  if(read_bits(b,1)) {
    fprintf(stderr, "Load inter quant matrix not implemented\n");
    for(i=0; i<16; i++)
      read_bits(b,32);
  }

#ifdef DEBUG    
  printf("Sequence header:\n");
  printf("Horiz: %i ", seq->horizontal);
  printf("Vert: %i ", seq->vertical);
  printf("Aspect Ration: %x ", seq->aspect);
  printf("Picture Rate: %x ", seq->pic_rate);
  printf("Bit Rate: %i ", seq->bit_rate);
  printf("VBV Buffer Size: %i ", seq->vbv_bufsize);
  printf("Constrained: %x ", seq->constrained);
#endif
  return(seq);
} 

int parse_GOP(bitstream *b) {
  time_code t;

  t.drop_frame_flag = read_bits(b, 1);
  t.hours = read_bits(b, 5);
  t.minutes = read_bits(b, 6);
  t.marker = read_bits(b, 1);
  t.seconds = read_bits(b, 6);
  t.pictures = read_bits(b, 6);

  read_bits(b,2);
  align_bits(b);
#ifdef DEBUG
  printf("GOP header: ");
  printf("%i %i %i %i Secs %i, Pics %i\n", t.drop_frame_flag, t.hours,
	 t.minutes,  t.marker, t.seconds, t.pictures);
#endif
  printf("\n");
  return(0);
}
     
int parse_picture(bitstream *b) {
  unsigned long int tmp;

  /********************************/
  //  yuv_blocks_to_rgb_buffer();
  /********************************/

  pic->tem_ref = read_bits(b, 10);
  pic->type = read_bits(b, 3);
  pic->vbv_delay = read_bits(b, 16);
  if((pic->type == 2) || (pic->type == 3)) {
    pic->fmv_full = read_bits(b, 1);
    pic->fmv_code = read_bits(b, 3);
  } else {
    pic->fmv_full = pic->fmv_code = 0;
  }
  if(pic->type == 3) {
    pic->bmv_full = read_bits(b, 1);
    pic->bmv_code = read_bits(b, 3);
  } else {
    pic->bmv_full = pic->bmv_code = 0;
  }
#ifdef DEBUG
  printf("Picture header: Ref %i, Type %i, Delay %i, FMV %i %i, BMV %i %i\n",
	 pic->tem_ref, pic->type, pic->vbv_delay, pic->fmv_full,
	 pic->fmv_code, pic->bmv_full, pic->bmv_code);
#endif
  printf("%c", ((pic->type == 1) ? 'I' : ((pic->type == 2) ? 'P' : 'B')));
  fflush(stdout);
  while((tmp = read_bits(b, 1)) == 1)
    printf("Pic extra info: %lx", read_bits(b,8));

  return(0);
}

void parse_slice(bitstream *b, u_char slicenum) {
  unsigned long int quant;

  /**************************/
#ifndef EX_BITSTREAM
  reset_reconstruct();
#endif

  quant = read_bits(b, 5);
  while(read_bits(b, 1))
    printf("Slice Extra Data: %lx\n", read_bits(b, 8));
#ifdef DEBUG
  printf("Slice %x, Quant = %i\n", slicenum, quant);
#endif

  mv_st->fmv.right = mv_st->fmv.down = 0;
  mv_st->bmv.right = mv_st->bmv.down = 0;

  while(peek_bits(b, 23) != 0)
    parse_macroblock(b, pic, quant);

#ifndef EX_BITSTREAM
  if((pic->type == 1) || (pic->type == 2))
    yuv_blocks_to_rgb_buffer();
#endif
}

void parse_userdata(bitstream *b) {
  fprintf(stderr, "Userdata header:\n");
  while(peek_bits(b,23) != 0)
    fprintf(stderr, "%c", (char)read_bits(b, 8));
}

void parse_extension(bitstream *b) {
  printf("Extension header (not implemented!):\n");
  while(peek_bits(b,23) != 0) read_bits(b,8);
}

#ifdef EX_BITSTREAM
int main(int argc, char **argv) {
  bitstream *b;
  u_char c;

  pic = (pic_header *)malloc(sizeof(pic_header));

  //    b=init_bitstream("/home/archives/garbage/LenoSpecial.mpg"); 
    b=init_bitstream(argv[1]); 
    //    b=init_bitstream("/dosc/windows/desktop/bb.mpg");
  init_mb();
  while(1) {
    c=find_start_code(b);
    switch(c) {
    case 0xb8: parse_GOP(b); break;
    case 0x00: parse_picture(b); break;
    case 0xb3: parse_sequence(b); break;
    case 0xb2: parse_userdata(b); break;
    case 0xb5: parse_extension(b); break;
    case 0xb7: {
      fprintf(stderr, "Sequence end, exiting\n");
      exit(0);
    }
    case 0xb4: {
      fprintf(stderr, "Error start code found, exiting\n");
      exit(0);
    } 
    case 0x01 ... 0xAF: parse_slice(b, c); break;
    default:
      fprintf(stderr, "Bogus start code!\n");
      exit(-1);
    }
  }
}

#endif
