#include "quantize.h"


const int INTRALUMIN_QUANT_TABLE[64]={
  8,  16, 19, 22, 26, 27, 29, 34,
  16, 16, 22, 24, 27, 29, 34, 37,
  19, 22, 26, 27, 29, 34, 34, 38,
  22, 22, 26, 27, 29, 34, 37, 40,
  22, 26, 27, 29, 32, 35, 40, 48,
  26, 27, 29, 32, 35, 40, 48, 58,
  26, 27, 29, 34, 38, 46, 56, 69,
  27, 29, 35, 38, 46, 56, 69, 83
};


void iquant_intra(int *coeffs, int blocknum){
  static int i;

  /* Should be dezigzaged before now... */
  for(i=1; i < 64; i++){
    coeffs[i] = (2 * coeffs[i] * state->quantizer_scale * INTRALUMIN_QUANT_TABLE[i])/16;
    
    if ((coeffs[i] & 1) == 0)
      coeffs[i] = coeffs[i] - ((coeffs[i] < 0) ? -1 : 1);
    
    if (coeffs[i] > 2047) 
      coeffs[i] = 2047;

    if (coeffs[i] < -2048)
      coeffs[i] = -2048;
  }
  
  coeffs[0] = coeffs[0] * 8;

  if (((state->macroblock_address - state->past_intra_address) > 1) &&
      (blocknum == 0))
    state->dct_dc_y_past = 128*8;

  if(blocknum < 4) {
    coeffs[0] = state->dct_dc_y_past + coeffs[0];
    state->dct_dc_y_past = coeffs[0];
  } else if(blocknum == 4) {
    if((state->macroblock_address - state->past_intra_address) > 1)
      state->dct_dc_cb_past = 128*8;
    coeffs[0] = state->dct_dc_cb_past + coeffs[0];
    state->dct_dc_cb_past = coeffs[0];
  } else {
    if((state->macroblock_address - state->past_intra_address) > 1)
      state->dct_dc_cr_past = 128*8;
    coeffs[0] = state->dct_dc_cr_past + coeffs[0];
    state->dct_dc_cr_past = coeffs[0];
  }
  return;
}

void iquant_inter(int *coeffs, int quant_scale) {
  int i, sign;

  for(i=0; i<64; i++) {
    sign = (coeffs[i] < 0) ? -1 : 1;
    if(coeffs[i] == 0) sign = 0;
    coeffs[i] = (2*coeffs[i] + sign)*(quant_scale * 16) / 16;
    if((coeffs[i] & 1) == 0)
      coeffs[i] -= sign;

    if (coeffs[i] > 2047) 
      coeffs[i] = 2047;

    if (coeffs[i] < -2048)
      coeffs[i] = -2048;
  }
}    

/********* Quantization functions **************/


void quant_intra(int *coeffs, int blocknum){
  static int i, sign, qvalue;

  /* Should not be zigzaged before now... */
  for(i=1; i < 64; i++){
    sign = (coeffs[i] < 0) ? -1 : 1;
    if(coeffs[i] == 0) sign = 0;

    qvalue = seq_iface_state->quantizer_scale * INTRALUMIN_QUANT_TABLE[i];
    coeffs[i] = (((coeffs[i] << 4) + sign*qvalue) / (qvalue << 1));
    if (i%8 > 3) coeffs[i] = 0;
  }
  coeffs[0] = (coeffs[0] + 4 * ((coeffs[0] < 0) ? -1 : 1)) / 8;

  /* Calculate DC differential */
  if(blocknum < 4) {
    coeffs[0] = coeffs[0] -   seq_iface_state->dct_dc_y_past;    
    seq_iface_state->dct_dc_y_past += coeffs[0];
  } else if(blocknum == 4) {
    coeffs[0] = coeffs[0] -   seq_iface_state->dct_dc_cb_past;
    seq_iface_state->dct_dc_cb_past += coeffs[0];
  } else {
    coeffs[0] = coeffs[0] -   seq_iface_state->dct_dc_cr_past;
    seq_iface_state->dct_dc_cr_past += coeffs[0];
  }

  return;
}

void quant_inter(int *coeffs) {
  int i;

  for(i=0; i<64; i++) {
    coeffs[i] = (coeffs[i] * 8) / (seq_iface_state->quantizer_scale * 16);
    if(coeffs[i] > 255) fprintf(stderr, "ACK too big!\n");
    if(coeffs[i] < -256) fprintf(stderr, "ACK too small!\n");
    if (i%8 > 3) coeffs[i] = 0;
  }
}
