#include "reconstruct.h"
#include <assert.h>

extern pic_header *pic;
frame_t scratch, past_ref, tmp_frame;

//#define DECODER

void init_reconstruct(seq_header *header){
  int i, j;

  state = (reconstruct_state *) malloc(sizeof(reconstruct_state));

  state->header = header;
  
  state->dct_dc_y_past = 128*8;  /* Default predictions ??? */
  state->dct_dc_cb_past = 128*8; /* ? */
  state->dct_dc_cr_past = 128*8; /* ? */
  state->quantizer_scale = 8; /* Will be changed on first macroblock*/

  state->past_intra_address = -1;
  state->macroblock_address = -1;

  state->hscan_length = (header->horizontal+15)/16;
  state->vscan_length = (header->vertical+15)/16;
  state->scan_length = state->hscan_length * state->vscan_length; 

  state->outputpels = (int ***) malloc(state->scan_length * sizeof(short ***));

  for(i = 0; i <= state->scan_length; i++)
    state->outputpels[i] = (int **) malloc(6 * sizeof(short **));

  for(i = 0; i <= state->scan_length; i++)
    for(j = 0; j < 6; j++)
      state->outputpels[i][j] = (int *) malloc(64 * sizeof(int));

  scratch.y = (int *) malloc(state->scan_length * sizeof(int) * 16 *16 * 2);
  scratch.u = (int *) malloc(state->scan_length * sizeof(int) / 4 * 16 * 16);
  scratch.v = (int *) malloc(state->scan_length * sizeof(int) / 4 * 16 * 16);
  past_ref.y = (int *) malloc(state->scan_length * sizeof(int) * 16 *16);
  past_ref.u = (int *) malloc(state->scan_length * sizeof(int) / 4 * 16 * 16);
  past_ref.v = (int *) malloc(state->scan_length * sizeof(int) / 4 * 16 * 16);

  return;
}

void reset_reconstruct() {
  state->dct_dc_y_past = 128*8;
  state->dct_dc_cb_past = 128*8;
  state->dct_dc_cr_past = 128*8;
}

void copyblock(int *src, int *dest) {
  int i;

  for(i=0; i<8; i++) {
    memcpy(dest, src, 8*sizeof(int));
    dest += 16*state->hscan_length;
    src +=  8;
  }
}


void motion_compensate(macroblock *mb, frame_t frame, frame_t ref) {
  int i, j, rowstride;
  int mbx, mby, baseptr;
  int src;
  int right_for, down_for, right_half_for, down_half_for;

  //  mbx = state->macroblock_address % state->hscan_length;
  //  mby = state->macroblock_address / state->hscan_length;
  mbx = mb->address % state->hscan_length;
  mby = mb->address / state->hscan_length;
  rowstride = 16*state->hscan_length;

  baseptr = 16*(mby*16*state->hscan_length + mbx);

  right_for = mb->fmv.right >> 1;
  right_half_for = mb->fmv.right & 0x01;
  down_for = mb->fmv.down >> 1;
  down_half_for = mb->fmv.down & 0x01;

  src = baseptr + (down_for*rowstride) + right_for;
//  printf(":%i %i ", right_for, down_for);

#if 1
//  assert(baseptr + 15*rowstride + 15 < 76800);
  for(i=0; i<16; i++)
    for(j=0; j<16; j++) {
      if((!right_half_for) && (!down_half_for))
	frame.y[baseptr+i*rowstride + j] += ref.y[src+i*rowstride + j];
      else if((!right_half_for) && (down_half_for))
	frame.y[baseptr+i*rowstride + j] +=
	  (ref.y[src+i*rowstride + j] + ref.y[src+(i+1)*rowstride + j]) / 2;
      else if((right_half_for) && (!down_half_for))
	frame.y[baseptr+i*rowstride + j] +=
	  (ref.y[src+i*rowstride + j] + ref.y[src+i*rowstride + 1 + j]) / 2;
      else
	frame.y[baseptr+i*rowstride + j] +=
	  (ref.y[src+i*rowstride + j] + ref.y[src+i*rowstride + 1 + j] +
	   ref.y[src+(i+1)*rowstride + j] +	ref.y[src+(i+1)*rowstride+1+j]) / 4;
    }
#endif
#if 0
  for(i=0; i<16; i++)
    for(j=0; j<16; j++)
      frame.y[baseptr+i*rowstride + j] += 100;
#endif


  
  rowstride = 8*state->hscan_length;
  baseptr = 8*(mby*8*state->hscan_length + mbx);

  right_for = (mb->fmv.right/2) >> 1;
  right_half_for = (mb->fmv.right/2) & 0x01;
  down_for = (mb->fmv.down/2) >> 1;
  down_half_for = (mb->fmv.down/2) & 0x01;

  src = baseptr + (down_for*rowstride) + right_for;
#if 1
  for(i=0; i<8; i++)
    for(j=0; j<8; j++) {
      if(!right_half_for && !down_half_for) {
	frame.u[baseptr+i*rowstride + j] += ref.u[src+i*rowstride + j];
	frame.v[baseptr+i*rowstride + j] += ref.v[src+i*rowstride + j];
      } else if(!right_half_for && down_half_for) {
	frame.u[baseptr+i*rowstride + j] +=
	  (ref.u[src+i*rowstride + j] + ref.u[src+(i+1)*rowstride + j]) / 2;
	frame.v[baseptr+i*rowstride + j] +=
	  (ref.v[src+i*rowstride + j] + ref.v[src+(i+1)*rowstride + j]) / 2;
      } else if(right_half_for && !down_half_for) {
	frame.u[baseptr+i*rowstride + j] +=
	  (ref.u[src+i*rowstride + j] + ref.u[src+i*rowstride + 1 + j]) / 2;
	frame.v[baseptr+i*rowstride + j] +=
	  (ref.v[src+i*rowstride + j] + ref.v[src+i*rowstride + 1 + j]) / 2;
      } else {
	frame.u[baseptr+i*rowstride + j] +=
	  (ref.u[src+i*rowstride + j] + ref.u[src+i*rowstride + 1 + j] +
	   ref.u[src+(i+1)*rowstride + j] + ref.u[src+(i+1)*rowstride+1+j])/ 4;
	frame.v[baseptr+i*rowstride + j] +=
	  (ref.v[src+i*rowstride + j] + ref.v[src+i*rowstride + 1 + j] +
	   ref.v[src+(i+1)*rowstride + j] + ref.v[src+(i+1)*rowstride+1+j])/ 4;
      }
    }
#else
  for(i=0; i<8; i++)
    for(j=0; j<8; j++)
      frame.u[baseptr+i*rowstride + j] += 128;

  for(i=0; i<8; i++)
    for(j=0; j<8; j++)
      frame.v[baseptr+i*rowstride + j] += 128;
#endif
}


void process_macroblock(macroblock *mb, frame_t frame, frame_t ref){
  int tmp_block[64];
  int mbx, mby, baseptr, i, baseptru;
  /******* ASSUMING ALL INTRA FRAMES!!!!!!! ******/
  
  state->quantizer_scale = mb->quant; /*Always right */

  state->past_intra_address = state->macroblock_address;
  state->macroblock_address = mb->address;


  mbx = state->macroblock_address % state->hscan_length;
  mby = state->macroblock_address / state->hscan_length;
//  printf("%c", (mb->type & MACROINTRA_MASK) ? 'I' : 'M');
  baseptr = 16*(mby*16*state->hscan_length + mbx);
  baseptru = 8*(mby*8*state->hscan_length + mbx);

  /* Corrupts given blocks!!! */
#if 0
  if(!(mb->type & MACROINTRA_MASK)) {
#endif
  for(i=0; i<6; i++) {
    dezigzag_block(mb->blocks[i]);
    if(mb->type & MACROINTRA_MASK)
      iquant_intra(mb->blocks[i], i);
    else
      iquant_inter(mb->blocks[i], mb->quant);
  }

  idct2d(mb->blocks[0], tmp_block);
  copyblock(tmp_block, &frame.y[baseptr]);
  idct2d(mb->blocks[1], tmp_block);
  copyblock(tmp_block, &frame.y[baseptr + 8]);
  idct2d(mb->blocks[2], tmp_block);
  copyblock(tmp_block, &frame.y[baseptr + 8*16*state->hscan_length]);
  idct2d(mb->blocks[3], tmp_block);
  copyblock(tmp_block, &frame.y[baseptr + 8*16*state->hscan_length + 8]);
  idct2d(mb->blocks[4], tmp_block);
  for(i=0; i<8; i++)
    memcpy(&frame.u[baseptru + i*8*state->hscan_length], 
    	   tmp_block + i*8, 8*sizeof(int));
  idct2d(mb->blocks[5], tmp_block);
  for(i=0; i<8; i++)
    memcpy(&frame.v[baseptru + i*8*state->hscan_length], 
	   &tmp_block[i*8], 8*sizeof(int));
#if 0
        }
  
#endif
  if(!(mb->type & MACROINTRA_MASK)) {
    motion_compensate(mb, frame, ref);
     reset_reconstruct();
  }
  /* YUV Data is now sitting in outputpels */ 

  return;
}

void skip_mblock(int num) {
  macroblock mb;
  int i;

  reset_reconstruct();
  mb.fmv.right = mb.fmv.down = mb.bmv.right = mb.bmv.down = 0;
  for(i=0; i<num; i++) {
    state->past_intra_address = state->macroblock_address;
    state->macroblock_address++;
    mb.address = state->macroblock_address;
    motion_compensate(&mb, scratch, past_ref);
  }
}

void yuv_blocks_to_rgb_buffer(){
  static int max_length;
  static int vpos_pixel, hpos_pixel;
  static int width, height;
  static int y, cb, cr;
  static int r, g, b;
  static int tmp;
  static int first=0;
  static int redshift=0, greenshift=0, blueshift=0;
  static int mask;
  static int px, py, y_lum, y_chrom;

  //      while(!getchar());
  if(!first) {
    first=1;
    mask = display_state->rmask;
    while(mask >>= 1) redshift++; redshift-=7;
    mask = display_state->gmask;
    while(mask >>= 1) greenshift++; greenshift-=7;
    mask = display_state->bmask;
    while(mask >>= 1) blueshift++; blueshift-=7;
    /*
    fprintf(stderr, "RGB masks: %lx %i %lx %i %lx %i\n", display_state->rmask,
	    redshift, display_state->gmask, greenshift,
	    display_state->bmask, blueshift);
	    */
  }

  width = state->header->horizontal;
  height = state->header->vertical;
  max_length = width * height;

  vpos_pixel = -1;
  hpos_pixel = -1;

  //  for(i = 0; i < max_length; i++){
  for(py=0; py < state->header->vertical; py++) {
    /* Calculate these once */
    y_lum = py*state->hscan_length*16;
    y_chrom = (py/2)*state->hscan_length*8;
    for(px=0; px < state->header->horizontal; px++) {
#if 0
    /* reverse map to appropriate YUVs */
    /* Divisions and mods by non powers of two are slow! */
    //    vpos_pixel = i / width;
    //    hpos_pixel = i % width;
    if(++hpos_pixel == width) {
      hpos_pixel = 0;
      vpos_pixel++;
    }

    blockset = hpos_pixel/16 + (vpos_pixel/16) * state->hscan_length;
    // blocklumpel = hpos_pixel % 8 + ((vpos_pixel % 8) * 8);
    blocklumpel = (hpos_pixel & 7) + ((vpos_pixel & 7) << 3);
    // blockcpel = ((hpos_pixel % 16) / 2) + ((vpos_pixel %16) / 2) * 8; 
    blockcpel = ((hpos_pixel & 15) >> 1) + ((vpos_pixel & 14) << 2); 

    if (((hpos_pixel % 16) < 8) && ((vpos_pixel % 16) < 8))
      y = state->outputpels[blockset][0][blocklumpel];

    else if (((hpos_pixel % 16) > 7) && ((vpos_pixel % 16) < 8))
      y = state->outputpels[blockset][1][blocklumpel];

    else if (((hpos_pixel % 16) < 8) && ((vpos_pixel % 16) > 7))
      y = state->outputpels[blockset][2][blocklumpel];

    else 
      y = state->outputpels[blockset][3][blocklumpel];

    cb = state->outputpels[blockset][4][blockcpel];
    cr = state->outputpels[blockset][5][blockcpel];
#endif
    y = scratch.y[y_lum + px];
    cb = scratch.u[y_chrom + (px >> 1)];
    cr = scratch.v[y_chrom + (px >> 1)];

    if(y > 255) y = 255;
    if(cb > 255) cb = 255;
    if(cr > 255) cr = 255;
    if(y < 0) y = 0;
    if(cr < 0) cr = 0;
    if(cb < 0) cb = 0;

    /* Convert to RGB... and put in buffer */
    tmp = 76290 * (y-16);
    cb -= 128;
    cr -= 128;

    r = (tmp - 118*cb + 104582*cr) >> 16;
    g = (tmp - 25651*cb - 53314*cr) >> 16;
    b = (tmp + 132239*cb - 79*cr) >> 16;

    if (r > 255) r = 255;
    else if (r < 0) r = 0;
    if (g > 255) g = 255;
    else if (g < 0) g = 0;
    if (b > 255) b = 255;
    else if (b < 0) b = 0;

        displaybuffer[y_lum+px] = ((r << redshift) | (g << greenshift) |
    			(b << blueshift));
	//    displaybuffer[y_lum+px] = (((unsigned char) r) | ((unsigned char )g << 8) | ((unsigned char)b << 16));
  }
  }
#ifdef DECODER
  tmp_frame = past_ref;
  past_ref = scratch;
  scratch = tmp_frame;

  memset(scratch.y, 0, sizeof(int)*state->scan_length*16*16);
  memset(scratch.u, 0, sizeof(int)*state->scan_length*8*8);
  memset(scratch.v, 0, sizeof(int)*state->scan_length*8*8);
#endif
}














