#include "seq_iface.h"
#include "errno.h"

extern frame_t scratch;
extern unsigned char *cframe;
frame_t mc;
unsigned char *past_ref;

#define GOP_SIZE 10
#define I_TYPE 1
#define P_TYPE 2
#define B_TYPE 3

//#define ALL_INTRA
//#define SIZE_X 176
//#define SIZE_Y 144
#define SIZE_X 352
#define SIZE_Y 288

/*
const int encode_order[GOP_SIZE] = {I_TYPE, B_TYPE, B_TYPE, P_TYPE,
				    B_TYPE, B_TYPE, P_TYPE,
				    B_TYPE, B_TYPE, P_TYPE};
*/

#ifndef ALL_INTRA
const int encode_order[GOP_SIZE] = {I_TYPE, P_TYPE, P_TYPE, P_TYPE,
				    P_TYPE, P_TYPE, P_TYPE,
				    P_TYPE, P_TYPE, P_TYPE};

#else
const int encode_order[GOP_SIZE] = {I_TYPE, I_TYPE, I_TYPE, I_TYPE,
				    I_TYPE, I_TYPE, I_TYPE,
				    I_TYPE, I_TYPE, I_TYPE};

#endif

const int SKIP_BLOCK[6] = {0xfe, 0xfd, 0xfb, 0xf7, 0xef, 0xdf};

#define SAT(x) (x<-256) ? -256 : ((x>255) ? 255 : x)

char_frame_t reorder[4];

void init_sequence_interface(char filename[], int width, int height, int qscale){
  int i;

  seq_iface_state = (seq_iface *) malloc(sizeof(seq_iface));

  seq_iface_state->YUV_sequence = fopen(filename, "rb");
  if (seq_iface_state->YUV_sequence == NULL) {
    perror("SEQ_IFACE: File not opened.");
    exit(-1);
  }

  seq_iface_state->width = width;
  seq_iface_state->height = height;

  seq_iface_state->scan_width = width / 16 + (((width %16) > 0) ? 1 : 0 );
  seq_iface_state->scan_height = height / 16 + (((height %16) > 0) ? 1 : 0 );
  seq_iface_state->scan_length =  seq_iface_state->scan_width *  seq_iface_state->scan_height;

  seq_iface_state->dct_dc_y_past = 128;// * 8;
  seq_iface_state->dct_dc_cb_past = 128;// * 8;
  seq_iface_state->dct_dc_cr_past = 128;// * 8;
  seq_iface_state->past_intra_address = -1;
  seq_iface_state->macroblock_address =  -1;
  seq_iface_state->quantizer_scale = qscale;

  seq_iface_state->Y_buffer = (unsigned char *) malloc(width * height * sizeof(unsigned char));
  seq_iface_state->U_buffer = (unsigned char *) malloc(((width * height) / 4) * sizeof(unsigned char));
  seq_iface_state->V_buffer = (unsigned char *) malloc(((width * height) / 4) * sizeof(unsigned char));

  seq_iface_state->usual_macroblock = (macroblock *) malloc(sizeof(macroblock));

  seq_iface_state->usual_macroblock->address = -1;
  seq_iface_state->usual_macroblock->quant = 8; /*?????? */
  seq_iface_state->usual_macroblock->type = MACROINTRA_MASK; /*****?????*****/
  seq_iface_state->usual_macroblock->cbp = 0x3f;
  seq_iface_state->usual_macroblock->blocks[0] = (int *) malloc(64 * sizeof(int));
  seq_iface_state->usual_macroblock->blocks[1] = (int *) malloc(64 * sizeof(int));
  seq_iface_state->usual_macroblock->blocks[2] = (int *) malloc(64 * sizeof(int));
  seq_iface_state->usual_macroblock->blocks[3] = (int *) malloc(64 * sizeof(int));
  seq_iface_state->usual_macroblock->blocks[4] = (int *) malloc(64 * sizeof(int));
  seq_iface_state->usual_macroblock->blocks[5] = (int *) malloc(64 * sizeof(int));

  i_ref.y = (int *)malloc(width*height*sizeof(int));
  i_ref.u = (int *)malloc(width*height*sizeof(int)/4);
  i_ref.v = (int *)malloc(width*height*sizeof(int)/4);
  p_ref.y = (int *)malloc(width*height*sizeof(int));
  p_ref.u = (int *)malloc(width*height*sizeof(int)/4);
  p_ref.v = (int *)malloc(width*height*sizeof(int)/4);
  mc.y = (int *)malloc(width*height*sizeof(int));
  mc.u = (int *)malloc(width*height*sizeof(int)/4);
  mc.v = (int *)malloc(width*height*sizeof(int)/4);

  for(i=0; i<4; i++) {
    reorder[i].y = (unsigned char *)malloc(width*height*sizeof(char));
    reorder[i].u = (unsigned char *)malloc(width*height*sizeof(char)/4);
    reorder[i].v = (unsigned char *)malloc(width*height*sizeof(char)/4);
  }
  seq_iface_state->head = seq_iface_state->tail = 0;
  past_ref = (unsigned char *) malloc(width*height*sizeof(char));

  return;
}

void set_quant(int quantize_level){
  seq_iface_state->usual_macroblock->quant = quantize_level;
  seq_iface_state->quantizer_scale = quantize_level;

  return;
}

void reset_slice() {
  seq_iface_state->dct_dc_y_past = 128;
  seq_iface_state->dct_dc_cb_past = 128;
  seq_iface_state->dct_dc_cr_past = 128;
  state->dct_dc_y_past = 128*8;  /* Default predictions ??? */
  state->dct_dc_cb_past = 128*8; /* ? */
  state->dct_dc_cr_past = 128*8; /* ? */

}

int read_frame(char_frame_t buffer) {
  if (fread(buffer.y, sizeof(unsigned char), 
	    seq_iface_state->height * seq_iface_state->width,
	    seq_iface_state->YUV_sequence) != 0){
    fread(buffer.u, sizeof(unsigned char), 
	  (seq_iface_state->height * seq_iface_state->width)/4,
	  seq_iface_state->YUV_sequence);
    fread(buffer.v, sizeof(unsigned char), 
	  (seq_iface_state->height * seq_iface_state->width)/4,
	  seq_iface_state->YUV_sequence);
    return(1);
  }
  else {
    return 0;
  }
}

int encode_frame(char_frame_t frame, int frame_type)
{
  static int temp_block[64];
  static int address, width, scan_width, base, ctr, ctr2;
  static int address_inc=1;
  int i, j, sum, allzero;
  int mbx, mby;
  frame_t tmpframe;
  motion_vec mv;

  mv.right = -16;
  mv.down = -16;

  reset_fmv_state();
  reset_bmv_state();

  seq_iface_state->usual_macroblock->address = -1;

  scan_width = seq_iface_state->scan_width;
  width = seq_iface_state->width;
  reset_slice();
  reset_reconstruct();

  if(frame_type != 1) {
//    for(i=0; i<SIZE_X*SIZE_Y; i++)
//      past_ref[i] = (unsigned char)i_ref.y[i];
  
    for(i=0; i<SIZE_X*SIZE_Y; i++)
      cframe[i] = (unsigned char) frame.y[i];
    calc_halfpels();
  }

  memset(mc.y, 0, SIZE_X*SIZE_Y*sizeof(int));
  memset(mc.u, 0, SIZE_X*SIZE_Y*sizeof(int)/4);
  memset(mc.v, 0, SIZE_X*SIZE_Y*sizeof(int)/4);

  for(address=0; address < seq_iface_state->scan_length; address++) {
  /* Bookeeping for addresses */
    seq_iface_state->past_intra_address =  address-1;
    seq_iface_state->usual_macroblock->address = address;
    seq_iface_state->usual_macroblock->quant = seq_iface_state->quantizer_scale;
    seq_iface_state->macroblock_address = address;
    mbx = address % seq_iface_state->scan_length;
    mby = address / seq_iface_state->scan_length;

    if(frame_type == 1)
      seq_iface_state->usual_macroblock->type = 16;
    else
      seq_iface_state->usual_macroblock->type = 10;
    seq_iface_state->usual_macroblock->cbp = 0x3f;
    /* Get the data from the input buffer into the macroblock structure... */
    if(frame_type == 1) {
    base = address/scan_width * 16 * width + address % scan_width * 16;

    for(ctr = 0; ctr < 8; ctr++) {
      for(ctr2 = 0; ctr2 < 8; ctr2++) {
	  seq_iface_state->usual_macroblock->blocks[0][ctr * 8 + ctr2] = 
	    (int) frame.y[base + ctr * width + ctr2];
	  
	  seq_iface_state->usual_macroblock->blocks[1][ctr * 8 + ctr2] = 
	    (int) frame.y[base + ctr * width + ctr2 + 8];
	}
    }
    for(ctr = 8; ctr < 16; ctr++) {
      for(ctr2 = 0; ctr2 < 8; ctr2++) {
	  seq_iface_state->usual_macroblock->blocks[2][(ctr-8) * 8 + ctr2] = 
	    (int) frame.y[base + ctr * width + ctr2];
	  seq_iface_state->usual_macroblock->blocks[3][(ctr-8) * 8 + ctr2] = 
	    (int) frame.y[base + ctr * width + ctr2 + 8];
      }
    }
  
    base = address/scan_width * 8 * (width / 2) + address % scan_width * 8;
  
    for(ctr = 0; ctr < 8; ctr++) {
      for(ctr2 = 0; ctr2 < 8; ctr2++) {
	seq_iface_state->usual_macroblock->blocks[4][ctr * 8 + ctr2] = 
	  (int) frame.u[base + ctr * (width/2) + ctr2];
	
	seq_iface_state->usual_macroblock->blocks[5][ctr * 8 + ctr2] = 
	  (int) frame.v[base + ctr * (width/2) + ctr2];
      }
    }
    } else {
      /* Inter-coding */
   //   block_search(mbx, mby, &mv);
      //      printf("MV: %i %i\n", mv.right, mv.down);

      seq_iface_state->usual_macroblock->fmv.right = 0; //mv.right;
      seq_iface_state->usual_macroblock->fmv.down = 0; //mv.down;
      motion_compensate(seq_iface_state->usual_macroblock, mc, i_ref);

    base = address/scan_width * 16 * width + address % scan_width * 16;

    for(ctr = 0; ctr < 8; ctr++) {
      for(ctr2 = 0; ctr2 < 8; ctr2++) {
	  seq_iface_state->usual_macroblock->blocks[0][ctr * 8 + ctr2] = 
	    SAT((int)(frame.y[base + ctr * width + ctr2] - 
		      mc.y[base + ctr * width + ctr2]));
	  
	  seq_iface_state->usual_macroblock->blocks[1][ctr * 8 + ctr2] = 
	    SAT((int)(frame.y[base + ctr * width + ctr2 + 8] -
		      mc.y[base + ctr * width + ctr2 + 8]));
	}
    }
    for(ctr = 8; ctr < 16; ctr++) {
      for(ctr2 = 0; ctr2 < 8; ctr2++) {
	  seq_iface_state->usual_macroblock->blocks[2][(ctr-8) * 8 + ctr2] = 
	    SAT((int)(frame.y[base + ctr * width + ctr2] -
		      mc.y[base + ctr * width + ctr2]));
	  seq_iface_state->usual_macroblock->blocks[3][(ctr-8) * 8 + ctr2] = 
	    SAT((int)(frame.y[base + ctr * width + ctr2 + 8] -
		      mc.y[base + ctr * width + ctr2 + 8]));
      }
    }
  
    base = address/scan_width * 8 * (width / 2) + address % scan_width * 8;
  
    for(ctr = 0; ctr < 8; ctr++) {
      for(ctr2 = 0; ctr2 < 8; ctr2++) {
	seq_iface_state->usual_macroblock->blocks[4][ctr * 8 + ctr2] = 
	  SAT((int)(frame.u[base + ctr * (width/2) + ctr2] -
		mc.u[base + ctr * (width/2) + ctr2]));
	
	seq_iface_state->usual_macroblock->blocks[5][ctr * 8 + ctr2] = 
	  SAT((int)(frame.v[base + ctr * (width/2) + ctr2] -
		mc.v[base + ctr * (width/2) + ctr2]));
      }
    }
    }
    /* Process the macroblock */

    for(i=0; i<6; i++) {
      memcpy(temp_block, seq_iface_state->usual_macroblock->blocks[i],
	     64 * sizeof(int));
      dct2d(temp_block, seq_iface_state->usual_macroblock->blocks[i]);

      if(seq_iface_state->usual_macroblock->type & MACROINTRA_MASK)
	quant_intra(seq_iface_state->usual_macroblock->blocks[i], i);
      else
	quant_inter(seq_iface_state->usual_macroblock->blocks[i]);

      zigzag_block(seq_iface_state->usual_macroblock->blocks[i]);

      if(!(seq_iface_state->usual_macroblock->type & MACROINTRA_MASK)) {
	allzero = 0;
	if(i==10)
	  seq_iface_state->usual_macroblock->blocks[i][63] = 1;
	for(j=0; j<64; j++) {
	  if (seq_iface_state->usual_macroblock->blocks[i][j] != 0)
	    allzero = 1;
	}
	if(!allzero) {
	  seq_iface_state->usual_macroblock->cbp &= SKIP_BLOCK[5-i];
	}
      }
    }
    if(seq_iface_state->usual_macroblock->cbp == 0) {
      seq_iface_state->usual_macroblock->type &= 0xf7;
      address_inc = 1;
    } else {
      if((mv.right == 0) && (mv.down == 0)) {
	seq_iface_state->usual_macroblock->type &= 0xfd;
	reset_fmv_state();
      }
      //	skip_mblock(1);
      //      	address_inc++;
      //	continue;
      //      }
    }

    //    seq_iface_state->usual_macroblock->address = address_inc;
    input_macro_b(seq_iface_state->usual_macroblock);
    //   seq_iface_state->usual_macroblock->address = address;

    if((frame_type == 1) || (frame_type == 2))
      process_macroblock(seq_iface_state->usual_macroblock, scratch, i_ref);

  }
  
  for(i=0; i<SIZE_X*SIZE_Y; i++)
    past_ref[i] = (unsigned char)scratch.y[i];
  
  yuv_blocks_to_rgb_buffer();
  
  update_image();
    
  tmpframe = i_ref;
  i_ref = scratch;
  scratch = tmpframe;

  return (1);
}

int main(int argc, char **argv)
{
  //  FILE *wfptr;
  macroblock *mb_p;
  int pic=-1;
  int qscale = 8;
  seq_header seq;

  if((argc != 5) && (argc != 6)) {
    fprintf(stderr, "Usage: encode filename x y outfile <quant>\n");
    exit(-1);
  }
  if(argc == 6) {
    qscale = atoi(argv[5]);
  }

  //  wfptr = fopen(argv[4], "wb");

  init_bitstream(argv[4], atoi(argv[2]), atoi(argv[3]), qscale);
  init_sequence_interface(argv[1], atoi(argv[2]), atoi(argv[3]), qscale);
  init_motion_est();
  seq.horizontal = atoi(argv[2]);
  seq.vertical = atoi(argv[3]);
//  init_reconstruct(atoi(argv[2]), atoi(argv[3]));
  init_reconstruct(&seq);

  display_state = display_init(seq.horizontal, seq.vertical);
  displaybuffer = (int*)malloc(seq.horizontal*seq.vertical*sizeof(int));
  *(display_state->data_ptr) = displaybuffer;
  fprintf(stderr, "Initialized Display...\n");
  
  while(read_frame(reorder[seq_iface_state->tail++])) {
    pic++;
    if(pic==10) {
      pic=0;
      write_GOP_header();
    }
    reorder[seq_iface_state->tail - 1].picref = pic;

    if(encode_order[pic] != 3) {
      write_new_frame(pic, encode_order[pic]);
      encode_frame(reorder[--seq_iface_state->tail], encode_order[pic]);
      while(seq_iface_state->tail > seq_iface_state->head) {
	write_new_frame(reorder[seq_iface_state->head].picref, 3);
	encode_frame(reorder[seq_iface_state->head++], encode_order[pic]);
      }
      seq_iface_state->head = seq_iface_state->tail = 0;
    }

  }

  end_bitstream();
  return(0);
}
