#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>
#include <stdlib.h>
#include "xdisplay.h"


typedef struct {
  Display *display;
  Visual *visual;
  Window window;
  int size_x;
  int size_y;
  int depth;
  GC gc;
  XImage *image;
  long eventmask;
} X_t;

typedef struct {
  int **buffer[30];
  int last;
  int first;
} frame_queue_t;

X_t *X;
frame_queue_t queue;
display_t *display_state;

/*
int handler(Display *d, XErrorEvent *e) {
  char ebuf[255];
  fprintf(stderr, XGetErrorText(d, e->error_code, ebuf, 255));
}
*/

Visual *
FindFullColorVisual (dpy, depth)
     Display *dpy;
     int *depth;
{
  XVisualInfo vinfo;
  XVisualInfo *vinfo_ret;
  int numitems, maxdepth;
  
  vinfo.class = TrueColor;
  
  vinfo_ret = XGetVisualInfo(dpy, VisualClassMask, &vinfo, &numitems);
  
  if (numitems == 0) return NULL;

  maxdepth = 0;
  while(numitems > 0) {
    if (vinfo_ret[numitems-1].depth > maxdepth) {
      maxdepth = vinfo_ret[numitems-1 ].depth;
    }
    numitems--;
  }
  XFree(vinfo_ret);

  if (maxdepth < 16) return NULL;

  if (XMatchVisualInfo(dpy, DefaultScreen(dpy), maxdepth, 
                       TrueColor, &vinfo)) {
    *depth = maxdepth;
    display_state->rmask = vinfo.red_mask;
    display_state->gmask = vinfo.green_mask;
    display_state->bmask = vinfo.blue_mask;
    return vinfo.visual;
  }
  
  return NULL;
}


display_t *display_init(int size_x, int size_y)
{
  Window root;
  long fgcolor, bgcolor;
  int screen;
  XGCValues gcval;
  int *data;
  int i, j;
  int depth;
  unsigned int mask;
  XSetWindowAttributes xswa;

  display_state = (display_t *) malloc(sizeof(display_t));

  X = (X_t *) malloc(sizeof(X_t));
  X->size_x = size_x;
  X->size_y = size_y;

  if(!(X->display = XOpenDisplay(":0.0"))) {
    perror("XOpenDisplay");
    exit(1);
  }
  //  XSetErrorHandler(&handler);

  screen = DefaultScreen(X->display);
  root = RootWindow(X->display,screen);

  X->visual = FindFullColorVisual(X->display, &depth);
  if(X->visual == NULL) {
    fprintf(stderr, "Can't find FullColor Visual.  Exiting.\n");
    exit(1);
  }
  X->depth = depth;

  fgcolor = BlackPixel(X->display, screen);
  bgcolor = WhitePixel(X->display, screen);

  /* Create Window */
  mask = CWBackPixel | CWColormap | CWBorderPixel;
  xswa.colormap = XCreateColormap(X->display, root, X->visual, AllocNone);
  xswa.background_pixel = fgcolor;
  xswa.border_pixel = bgcolor;

  X->window = XCreateWindow(X->display, root, 0, 0, size_x, size_y, 1,
			    depth, InputOutput, X->visual, mask, &xswa);
  if(X->window == 0) {
    perror("XCreateWindow");
    exit(1);
  } 

  /* Create Graphics Context */
  gcval.foreground = fgcolor;
  gcval.background = bgcolor;
  X->gc = XCreateGC(X->display, X->window, GCForeground|GCBackground, &gcval);


  X->eventmask = ButtonPressMask|ExposureMask|KeyPressMask;
  XSelectInput(X->display, X->window, X->eventmask);
  XMapWindow(X->display,X->window);

  data = (int *)malloc(size_x*size_y*sizeof(int));

  for(i=0; i<size_x; i++)
    for(j=0; j<size_y; j++)
      data[i+size_x*j] = 256*i + j;

  X->image=XCreateImage(X->display, X->visual, depth, ZPixmap, 0, (char *)data,
		     size_x, size_y, 32, 0);
  if(X->image == NULL) {
    perror("XCreateImage");
    exit(1);
  } 
  /*  queue.first = queue.last = 0; */
  display_state->data_ptr = (int **)&(X->image->data);
  return(display_state);
}

void update_image() {
      XPutImage(X->display, X->window, X->gc, X->image,
		0, 0, 0, 0, X->size_x, X->size_y);
}  

int handle_xevents() {
  XEvent event;
  int count=0;
  
  while((XCheckMaskEvent(X->display, X->eventmask, &event) != False) 
	&& (count < 10)) {
    count++;
    switch (event.type) {
    case Expose:
      update_image();
      break;
    case ButtonPress:
      update_image();
      break;
    case KeyPress:
      exit(0);
    default:
      fprintf(stderr,"Unexpected event: %d\n",event.type);
    }
  }
  return(0);
}

/* NOT USED ***********************************
int enqueue_frame(int *data) {
  int tmp;

  queue.last++;
  if(queue.last > 30) queue.last = 0;
  if(queue.last==queue.first) {
    fprintf(stderr, "Buffer overrun!!!\n");
    queue.last--;
    return(-1);
  }
  queue.buffer[queue.last] = data;

  tmp = queue.last - queue.first;
  return((tmp > 0) ? tmp : tmp + 30);
}

int advance_frame() {
  int tmp;

  queue.first++;
  if(queue.first == queue.last) {
    fprintf(stderr, "Buffer underrun!!!\n");
    queue.first--;
    return(-1);
  }
  free(X->image->data);
  X->image->data = queue.buffer[queue.first];
  update_image();

  tmp = queue.last - queue.first;
  return((tmp > 0) ? tmp : tmp + 30);
}
*/

#if 0
int main() {
  int i;
  int *data;
  int color=0;

  data=display_init(200, 400);
  while(1) {
    for(i=0; i<1000000; i++);
    /*      for(i=0; i<X->size_x; i++)
	    for(j=0; j<X->size_y; j++)
	data[i + (X->size_x)*j] = 256*256*i + 256*j; */
    fprintf(stderr, ".");
    for(i=0; i<200*400; i++)
      data[i] = color + i*0x00050302/400;
    update_image();
    handle_xevents();
    color += 0x00101010;
  }
}
#endif
